//+------------------------------------------------------------------+
//|                                           CustPrevCalculated.mqh |
//|                                         Copyright 2011, AlexSTAL |
//|                                           http://www.alexstal.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, AlexSTAL"
#property link      "http://www.alexstal.ru"

//+------------------------------------------------------------------+
//|                      |
//+------------------------------------------------------------------+
enum CPCPrepareDataResultCode
  {
   CPCPDRC_NoData,                     // ,      (  )    
   CPCPDRC_FullInitialization,         //     
   CPCPDRC_Synch,                      //       
   CPCPDRC_SynchOnlyLastBar,           //       (   )
   CPCPDRC_NoRecountNotRequired        //   ,     
  };

//+------------------------------------------------------------------+
//|                                    |
//|     HistorySynchSecond                     |
//+------------------------------------------------------------------+
enum CPCHistorySynchMode
  {
   CPCHSM_NotSynch,                    //  
   CPCHSM_Normal,                      //      
   CPCHSM_Paranoid                     //     
  };

//+------------------------------------------------------------------+
//| :       |
//+------------------------------------------------------------------+
enum CPCCACMR
  {
   CPCCACMR_CompareError,              //      
   CPCCACMR_Equal,                     //  
   CPCCACMR_Different                  //  
  };

//+------------------------------------------------------------------+
//|            |
//+------------------------------------------------------------------+
const uchar CPCH_time = 0;
const uchar CPCH_open = 1;
const uchar CPCH_high = 2;
const uchar CPCH_low = 4;
const uchar CPCH_close = 8;
const uchar CPCH_tick_volume = 16;
const uchar CPCH_spread = 32;
const uchar CPCH_real_volume = 64;


//+------------------------------------------------------------------+
//|  CCustPrevCalculated   prev_calculated         |
//+------------------------------------------------------------------+
class CCustPrevCalculated
  {
protected:
   //  
   bool Initialized;                         //         
   string Instrument;                        //  
   ENUM_TIMEFRAMES TimeFrame;                //  
   int BarsLimit;                            //    ,     
   //  
   uchar HistorySynch;                       //   //  
   CPCHistorySynchMode HistorySynchMode;     //           
   uint HistorySynchSecond;                  //   ( ) ,      , 0 -  
   //    ,      ,  
   uint DataBarsCount;                       //        ,     BarsLimit
   uint DataBarsCalculated;                  //     
   uint DataStartInd;                        //        (  )
   MqlRates Data[];                          //     . !  !
   //   
   bool DataBarsCuttingLeft;                 //      PrepareData()    
private:
   //  
   //----------------------------
   MqlRates tmpRates[];                                                                      //      
   CPCCACMR _CompareAndChangeMqlRates(MqlRates &MqlRatesTarget, MqlRates &MqlRatesSource);   //      HistorySynch
   int _CompareHistory(int LastBarInd, datetime TimeCompare, bool &FullInitialization);      //     
   //       
   uint PInd(uint I, bool AsSeries) {return( AsSeries?((DataStartInd+DataBarsCount-1-I) % BarsLimit):((DataStartInd+I) % BarsLimit) );}
public:
   // 
   void CCustPrevCalculated() {Initialized = false;}
   bool ItIsInitialized() {return(Initialized);}
   //    (  1    )
   bool InitData(string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute);
   //    
   CPCPrepareDataResultCode PrepareData();
   //     
   uint GetDataBarsCount() {return(DataBarsCount);}                              //    
   uint GetDataBarsCalculated() {return(DataBarsCalculated);}                    //     
   uint GetDataStartInd() {return(DataStartInd);}                                //     
   bool GetDataBarsCuttingLeft() {return(DataBarsCuttingLeft);}                  //     
      // :          
   double GetDataOpen(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].open:Data[(DataStartInd+shift) % BarsLimit].open);}
   double GetDataHigh(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].high:Data[(DataStartInd+shift) % BarsLimit].high);}
   double GetDataLow(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].low:Data[(DataStartInd+shift) % BarsLimit].low);}
   double GetDataClose(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].close:Data[(DataStartInd+shift) % BarsLimit].close);}
   datetime GetDataTime(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].time:Data[(DataStartInd+shift) % BarsLimit].time);}
   long GetDataTick_volume(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].tick_volume:Data[(DataStartInd+shift) % BarsLimit].tick_volume);}
   long GetDataReal_volume(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].real_volume:Data[(DataStartInd+shift) % BarsLimit].real_volume);}
   int GetDataSpread(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].spread:Data[(DataStartInd+shift) % BarsLimit].spread);}
  };
//+------------------------------------------------------------------+
bool CCustPrevCalculated::InitData(string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute)
{
   int tmpLimit = Limit;
   if (tmpLimit > TerminalInfoInteger(TERMINAL_MAXBARS))
      tmpLimit = TerminalInfoInteger(TERMINAL_MAXBARS);
   if (tmpLimit < 2)
      tmpLimit = 2;
   
   CPCHistorySynchMode tmpHSM = HSM;
   if (MQL5InfoInteger(MQL5_TESTING))
      tmpHSM = CPCHSM_NotSynch;
   
   uchar tmpHS = HS;
   if (tmpHS == 0)
      tmpHS = CPCH_open|CPCH_high|CPCH_low|CPCH_close|CPCH_tick_volume|CPCH_spread|CPCH_real_volume;
   
   uint tmpHSMinute = HSMinute * 60;
   
   if (!Initialized)
     {
      BarsLimit = tmpLimit;
      if (ArrayResize(Data, BarsLimit) == -1)
         return(false);
      
      Instrument = Instr;
      TimeFrame = TF;
      HistorySynchMode = tmpHSM;
      HistorySynchSecond = tmpHSMinute;
      
      DataBarsCount = 0; //        
      
      HistorySynch = tmpHS;
      
      Initialized = true;
      return(true);
     } else {
      //  ,    
      bool res = true;
      if (Instrument != Instr) res = false;
      if (TimeFrame != TF) res = false;
      if (BarsLimit != tmpLimit) res = false;
      if (HistorySynchMode != tmpHSM) res = false;
      if (HistorySynch != tmpHS) res = false;
      if (HistorySynchSecond != tmpHSMinute) res = false;
      return(res);
     }
}
//+------------------------------------------------------------------+
CPCCACMR CCustPrevCalculated::_CompareAndChangeMqlRates(MqlRates &MqlRatesTarget, MqlRates &MqlRatesSource)
{
   if (MqlRatesTarget.time != MqlRatesSource.time)
      return(CPCCACMR_CompareError);
   
   CPCCACMR res = CPCCACMR_Equal;
   
   if ( (HistorySynch & CPCH_open) != 0)
      if (MqlRatesTarget.open != MqlRatesSource.open)
        { MqlRatesTarget.open = MqlRatesSource.open; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_high) != 0)
      if (MqlRatesTarget.high != MqlRatesSource.high)
        { MqlRatesTarget.high = MqlRatesSource.high; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_low) != 0)
      if (MqlRatesTarget.low != MqlRatesSource.low)
        { MqlRatesTarget.low = MqlRatesSource.low; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_close) != 0)
      if (MqlRatesTarget.close != MqlRatesSource.close)
        { MqlRatesTarget.close = MqlRatesSource.close; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_tick_volume) != 0)
      if (MqlRatesTarget.tick_volume != MqlRatesSource.tick_volume)
        { MqlRatesTarget.tick_volume = MqlRatesSource.tick_volume; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_spread) != 0)
      if (MqlRatesTarget.spread != MqlRatesSource.spread)
        { MqlRatesTarget.spread = MqlRatesSource.spread; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_real_volume) != 0)
      if (MqlRatesTarget.real_volume != MqlRatesSource.real_volume)
        { MqlRatesTarget.real_volume = MqlRatesSource.real_volume; res = CPCCACMR_Different; }

   return(res);
}
//+------------------------------------------------------------------+
int CCustPrevCalculated::_CompareHistory(int LastBarInd, datetime TimeCompare, bool &FullInitialization)
{
   int res = -1;
   uint tmpHSS = HistorySynchSecond;
   for (int i = LastBarInd; i >= 0; i--)
     {
      //       
      //  ,   
      if (tmpHSS != 0)
         if ( (tmpRates[i].time + HistorySynchSecond) < TimeCompare)
            break;
      switch(_CompareAndChangeMqlRates(Data[PInd(i, false)], tmpRates[i]))
        {
         case CPCCACMR_Different:
            res = (res == -1)?i:MathMin(res, i);
            tmpHSS = 0; //   ,    
            break;
         case CPCCACMR_CompareError:
            FullInitialization = true;
            return(-1);
            break;
        }
     }
   return(res);
}
//+------------------------------------------------------------------+
CPCPrepareDataResultCode CCustPrevCalculated::PrepareData()
{
   if (!Initialized)
      return(CPCPDRC_NoData);
      
   int i;
   int BarsCopy = CopyRates(Instrument, TimeFrame, 0, BarsLimit, tmpRates);
   if (BarsCopy <= 0)
      return(CPCPDRC_NoData);
   //    tmpRates      BarsCopy 
   //     ,     -  
   //    
   //  0   ,  (BarsCopy-1)  
   
   //   
   datetime DataFirstBarTime = 0;                              //  
   datetime DataLastBarTime = 0;                               //  
   datetime tmpRatesFirstBarTime = tmpRates[0].time;           //  
   datetime tmpRatesLastBarTime = tmpRates[BarsCopy-1].time;   //  
   
   //     -  ,   ..
   bool FullInitialization = false;
   
   //  DataBarsCount = 0,    
   bool tmpInsertBars = false;
   if (DataBarsCount != 0)
     {
      DataFirstBarTime = Data[PInd(0, false)].time;              //  
      DataLastBarTime = Data[PInd(DataBarsCount-1, false)].time; //  
      //          
      int BarsOldTest = Bars(Instrument, TimeFrame, DataFirstBarTime, DataLastBarTime);
      if (BarsOldTest == 0)
         return(CPCPDRC_NoData);
      if (BarsOldTest == DataBarsCount)
        {
         if (tmpRatesFirstBarTime >= DataLastBarTime)  //   
            FullInitialization = true;
          if (tmpRatesFirstBarTime < DataFirstBarTime) //   
            FullInitialization = true;
        } else {
         //      
         tmpInsertBars = true;
         //FullInitialization = true; //   
        }
     } else
      FullInitialization = true;
   
   //   DataLastBarTime   
   int SearchIndLB = -1;
   if (!FullInitialization)
     {
      for (i = (BarsCopy-1); i >= 0; i--)
         if (tmpRates[i].time == DataLastBarTime)
           { SearchIndLB = i; break; }
      if (SearchIndLB == -1)
         FullInitialization = true;
     }
   
   //   tmpRatesFirstBarTime   
   int SearchIndRtime0 = -1; //       ,    
   if ( (!FullInitialization) && (tmpRatesFirstBarTime > DataFirstBarTime) ) //   -  
     {
      for (i = 0; i < (int)DataBarsCount; i++)
        {
         if (Data[PInd(i, false)].time == tmpRatesFirstBarTime)
           { SearchIndRtime0 = i; break; }
        }
      if (SearchIndRtime0 == -1)
         FullInitialization = true;
     }
   
   CPCPrepareDataResultCode tmpCPCPDRC = CPCPDRC_Synch;
   if (!FullInitialization)
     {
      //   -  
      DataBarsCuttingLeft = false;
      if (SearchIndRtime0 != -1)
        { //   ,        
         DataBarsCount = DataBarsCount - SearchIndRtime0;
         DataStartInd = PInd(SearchIndRtime0, false);
         DataFirstBarTime = tmpRatesFirstBarTime;
         DataBarsCuttingLeft = true;
        }
      
      if (tmpInsertBars)
        {
         int SearchIndInsBars = -1;
         for (i = 0; i < (int)DataBarsCount; i++)
            if (Data[PInd(i, false)].time != tmpRates[i].time)
              {
               SearchIndInsBars = i;
               break;
              }
         if (SearchIndInsBars > 0)
           {
            DataBarsCount = SearchIndInsBars;
            SearchIndLB = SearchIndInsBars - 1;
           } else {
            FullInitialization = true;
           }
        }
     }
     
   DataBarsCalculated = SearchIndLB;
   //        !
   if (!FullInitialization)
     {
      uint NewBars = BarsCopy-SearchIndLB-1;
      int tmpCH;
      if (NewBars > 0)
        {
         //   DataLastBarTime     
         uint BarsToCopy = NewBars + 1;
         DataBarsCount = DataBarsCount - 1;
         
         //      
         uint RightFreeBars = (BarsLimit - 1) - PInd(DataBarsCount - 1, false);
         DataBarsCount = DataBarsCount + BarsToCopy;
         uint ToCopyRight = MathMin(BarsToCopy, RightFreeBars);
         if (RightFreeBars != 0)
           {
            ArrayCopy(Data, tmpRates, BarsLimit - RightFreeBars, SearchIndLB, ToCopyRight);
            BarsToCopy = BarsToCopy - ToCopyRight;
           } else
            ToCopyRight = 0;
         if (BarsToCopy > 0)
            ArrayCopy(Data, tmpRates, 0, SearchIndLB + ToCopyRight, BarsToCopy);
         DataLastBarTime = tmpRatesLastBarTime;
         
         //   (  )
         if (HistorySynchMode != CPCHSM_NotSynch)
            if (SearchIndLB >= 1)
              {
               tmpCH = _CompareHistory(SearchIndLB - 1, TimeCurrent() - (NewBars * PeriodSeconds(TimeFrame)), FullInitialization);
               if (tmpCH != -1)
                  DataBarsCalculated = (tmpCH==0)?0:tmpCH - 1;
              }
        } else {
         //    
         //--------------------------
         //     
         switch(_CompareAndChangeMqlRates(Data[PInd(DataBarsCount - 1, false)], tmpRates[BarsCopy-1]))
           {
            case CPCCACMR_Equal:
               DataBarsCalculated = DataBarsCalculated + 1;
               tmpCPCPDRC = CPCPDRC_NoRecountNotRequired;
               break;
            case CPCCACMR_Different:
               tmpCPCPDRC = CPCPDRC_SynchOnlyLastBar;
               break;
            case CPCCACMR_CompareError:
               FullInitialization = true;
               break;
           }
         
         if (!FullInitialization)
           {
            //   (  )
            if (HistorySynchMode == CPCHSM_Paranoid)
               if (SearchIndLB >= 1)
                 {
                  tmpCH = _CompareHistory(SearchIndLB - 1, TimeCurrent(), FullInitialization);
                  if (tmpCH != -1)
                    {
                     DataBarsCalculated = (tmpCH==0)?0:tmpCH - 1;
                     tmpCPCPDRC = CPCPDRC_Synch;
                    }
                 }
           }
        } // if (NewBars > 0)
     }

   if (FullInitialization)
     {
      DataBarsCount = BarsCopy;
      DataStartInd = 0;
      DataBarsCalculated = 0;
      ArrayCopy(Data, tmpRates, 0, 0, DataBarsCount);
      tmpCPCPDRC = CPCPDRC_FullInitialization;
     }
   return(tmpCPCPDRC);
}